/////////////////////////////////////////////////////////////////////////////
// Ac97.cpp : SBLive! Codec Class Implementation File
// <THIS IS OPEN SOURCE CODE>
//
// v2.0 (01/07/2000)
// C++ Version
//
// <uses system file EMUAPS.VXD>
//
// History
// -------
//
// Date         Author          Description
// ----         ------          -----------
// 01/07/2000   Music Master    Initial VXD Version
//
//
// [Notes]
// (a) DEST_MASTER & DEST_PCM appear to combine to form 'Analog Out 1 & 2'
// (b) DEST volume operations appear to have NO effect on the output
//     of 'Analog 7 & 8'
// (c) Setting the source (SelectRecordIn function) sets what input is passed
//     from the codec to BOTH 'Analog In 1' and 'Analog In 2'.
// (d) Possibly still a bug that makes the Sblive_writeac97 function
//     time-out.

#include <stdio.h>

#include "8010.h"
#include "Ac97.h"

/////////////////////////////////////////////////////////////////////////////
// Initialise statics

HANDLE           Ac97::m_Sb_hw = NULL;
const int	     Ac97::m_Version = 0x0200; // (v2.0)

// output mapping array
static char dest_mapped[ DEST_END ] =
{
	AC97_MASTERVOLUME,
	AC97_HEADPHONEVOLUME,
	AC97_MASTERVOLUMEMONO,
	AC97_MASTERTONE,
	AC97_PCBEEPVOLUME,
	AC97_PHONEVOLUME,
	AC97_MICVOLUME,
	AC97_LINEINVOLUME,
	AC97_CDVOLUME,
	AC97_VIDEOVOLUME,
	AC97_AUXVOLUME,
	AC97_PCMOUTVOLUME
};

// data size mapping array
static char size_mapped[ 3 ] =
{
	1,	// 1 byte	== BYTE
	2,	// 2 bytes	== WORD
	4	// 4 bytes	== LONG
};

/////////////////////////////////////////////////////////////////////////////
// GetLibVersion : Ask for the library version

int Ac97::GetLibVersion( int part )
{
	// major of minor?
	if( part == LIBVER_MAJOR )
	{
		return( (m_Version & 0xFF00) >> 8 );
	}
	else
	{
		return( m_Version & 0xFF );
	}

}

/////////////////////////////////////////////////////////////////////////////
// Status : Handler working/setup OK?

bool Ac97::Status( void )
{
	// return status boolean
	if( m_Sb_hw )
	{
		return true;
	}
	else
	{
		return false;
	}
}

/////////////////////////////////////////////////////////////////////////////
// DetectSBLive : Get a handle to use the EMUAPS.VXD device with
//
// return = false-failed / true-handle set

bool Ac97::DetectSBLive( void )
{
	HANDLE devHandle;

	// open the VXD
	devHandle = CreateFile( EMU_VXD, 0, 0, 0, CREATE_NEW,
							FILE_FLAG_DELETE_ON_CLOSE, 0 );

	// open NOT OK?
	if( devHandle == INVALID_HANDLE_VALUE )
	{
		return false;
	}

	// save handle
	m_Sb_hw = devHandle;

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Reset : Reset AC97 device

bool Ac97::Reset( void )
{
	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return false;
	}


	// Disable Audio
	Vxd_Write( HCFG, VXD_LONG_DATA, HCFG_GPINPUT0 | HCFG_AUTOMUTE );

    // Enable the AUD bit
    Sblive_rmwac97( DEST_MASTER, 0x8000, 0x8000 );

    // Reset Ac97
    Sblive_writeac97( AC97_RESET, 0 );

    // Set Master Volume & PCM Volume (combines to make analogue output 1&2) to full
    SetVolumeOut( DEST_MASTER, 63 );
    SetVolumeOut( DEST_PCM, 63 );

	// Set default recording source to LINE with Zero Gain
	SetRecordSource( SOURCE_LINE );
	SetRecordGain( 0 );

	// Enable Audio
	Vxd_Write( HCFG, VXD_LONG_DATA, HCFG_GPINPUT0 | HCFG_AUTOMUTE | HCFG_MUTEBUTTONENABLE | HCFG_AUDIOENABLE );

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// SetVolumeOut : Set a new volume for a destination
//
// src = dest[ enum ]
// vol = 0 - 63 (max output)

bool Ac97::SetVolumeOut( int dest, int vol )
{
	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return false;
	}

	// volume and dest must be in ranage
	if( vol > MAX_OUT_VOLUME || dest >= DEST_END )
	{
		return false;
	}

	// invert vol
	vol = 0x3F - vol;

	// set volume for this dest
	Sblive_writeac97( dest_mapped[ dest ], 0x0101 * vol );

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// GetRecordGain : Get the current input gain
//
// return = 0 - 15 (max gain)

int Ac97::GetRecordGain( void )
{
	short tmpShort;

	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return INT_ERROR;
	}

	// read current gain
	Sblive_readac97( AC97_RECORDGAIN, &tmpShort );

	// return result
	return (int) tmpShort;
}

/////////////////////////////////////////////////////////////////////////////
// SetRecordGain : Set a new input gain for the current record source
//
// gain = 0 - 15 (max gain)

bool Ac97::SetRecordGain( int gain )
{
	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return false;
	}

	// gain must be in range
	if( gain > MAX_IN_GAIN )
	{
		return false;
	}

	// set gain for record
	Sblive_writeac97( AC97_RECORDGAIN, 0x0101 * gain );

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// SetRecordSource : Set a new record source
//
// src = gain[ enum ]

bool Ac97::SetRecordSource( int src )
{
	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return false;
	}

	// src must be in ranage
	if( src >= DEST_END )
	{
		return false;
	}

	// set record source
	Sblive_writeac97( AC97_RECORDSELECT, 0x0101 * src );

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// GetRecordSource : Return the enum ID of the current record source
//
// return = gain[ enum ]

int Ac97::GetRecordSource( void )
{
	short tmpShort;

	// Fail if no device handle set
	if( !m_Sb_hw )
	{
		return INT_ERROR;
	}

	// get record source
	Sblive_readac97( AC97_RECORDSELECT, &tmpShort );

	// return result
	return (int) tmpShort;
}

/////////////////////////////////////////////////////////////////////////////
// Sblive_readac97 : Reada word from the ac97

bool Ac97::Sblive_readac97( char index, short *pdata )
{
	bool bResult;

	// set the address
	bResult = Vxd_Write( AC97ADDRESS, VXD_BYTE_DATA, index );

	// read the data
	bResult &= Vxd_Read( AC97DATA, VXD_WORD_DATA, pdata );

	// return success code
	if( bResult )
	{
		return true;
	}
	else
	{
		return false;
	}

}

/////////////////////////////////////////////////////////////////////////////
// Sblive_writeac97 : Write a word to the ac97

bool Ac97::Sblive_writeac97( char index, short data )
{
    int iTimeout = 0;
    char cByte = 0;
	bool bResult = true;

    // if device has been cleared, return in error
    if( !m_Sb_hw )
    {
        return false;
    }

    // retry loop
    while( iTimeout < VXD_MAX_TIMEOUT && !cByte )
    {
    	// set the address
	    bResult = Vxd_Write( AC97ADDRESS, VXD_BYTE_DATA, index );

        // read byte
	    bResult &= Vxd_Read( AC97ADDRESS, VXD_BYTE_DATA, &cByte );

        iTimeout++;
    }

    // write timeout error?
    if( iTimeout >= VXD_MAX_TIMEOUT )
    {
        // write mesagebox (no window attached)
        MessageBox( NULL, TIMEOUT_ERROR_TEXT, "AC97 VXD Handler", MB_OK | MB_ICONEXCLAMATION );

        // clear device handle, as found to be invalid!
        m_Sb_hw = NULL;
    }

    // set the data
	bResult &= Vxd_Write( AC97DATA, VXD_WORD_DATA, data );

	// return success code
	if( bResult )
	{
		return true;
	}
	else
	{
		return false;
	}

}

/////////////////////////////////////////////////////////////////////////////
// Vxd_Write : Write data to the device via the EMUAPS.VXD

bool Ac97::Vxd_Write( int reg, int sizeID, long value )
{
	BOOL bResult;
	char InputBuffer[ 22 ];
	unsigned long BytesReturned;

	// make pointers
	short *wptr = (short *) &InputBuffer[0];
	long *lptr = (long *) (wptr+9);

	// clear memory and then set required values
	memset( &InputBuffer, 0, sizeof( InputBuffer ) );

	*(wptr) = 0xA4;			// <unknown magic number meaning>
	*(wptr+5) = 0x11A1;		// <unknown magic number meaning>
	*(wptr+1) = VXD_WRITE;	// write mode
	*(wptr+7) = sizeID;		// data size (VXD_LONG/WORD/BYTE_DATA)
	*(wptr+8) = reg;		// hardware register
	*lptr = value;			// sending value

	// call the vxd
	bResult = DeviceIoControl( m_Sb_hw, EMU_CONTROL_CODE,
							   &InputBuffer, sizeof( InputBuffer ),
							   0, 0,
							   &BytesReturned, 0 );

	// return success code
	if( bResult )
		return true;
	else
		return false;

}

/////////////////////////////////////////////////////////////////////////////
// Vxd_Read : Read data from the device via the EMUAPS.VXD

bool Ac97::Vxd_Read( int reg, int sizeID, void *value )
{
	BOOL bResult;
	char InputBuffer[ 22 ];
	char OutputBuffer[ 4 ];
	unsigned long BytesReturned;

	// make pointers
	short *wptr = (short *) &InputBuffer[0];
	long *lptr = (long *) (wptr+9);

	// clear memory and then set required values
	memset( &InputBuffer, 0, sizeof( InputBuffer ) );
	memset( &OutputBuffer, 0xFF, sizeof( OutputBuffer ) );

	*(wptr) = 0xA4;			// <unknown magic number meaning>
	*(wptr+5) = 0x11A1;		// <unknown magic number meaning>
	*(wptr+1) = VXD_READ;	// write mode
	*(wptr+7) = sizeID;		// data size (VXD_LONG/WORD/BYTE_DATA)
	*(wptr+8) = reg;		// hardware register
	*lptr = 0xFFFFFFFF; 	// <bytes unused for reading>

	// call the vxd
	bResult = DeviceIoControl( m_Sb_hw, EMU_CONTROL_CODE,
							   &InputBuffer, sizeof( InputBuffer ),
							   &OutputBuffer, size_mapped[ sizeID ],
							   &BytesReturned, 0 );

	// assign result (based on data size)
	switch( sizeID )
	{
	case VXD_BYTE_DATA:
		*((char *) value) = OutputBuffer[0];
		break;
	case VXD_WORD_DATA:
		*((short *) value) = OutputBuffer[0];
		break;
	case VXD_LONG_DATA:
		*((long *) value) = OutputBuffer[0];
		break;
	}

	// return success code
	if( bResult )
		return true;
	else
		return false;

}

/////////////////////////////////////////////////////////////////////////////
// Sblive_writeac97 : Write a word to the ac97 (masked)
//
// Exp: Word of data is read from the specified index, bits of this data
//      specified in the mask are cleared.  Bits of the sent data NOT
//      specified in the mask are cleared.  The remaining bits of both values
//      are OR-ed together.

bool Ac97::Sblive_rmwac97( char index, short data, short mask )
{
	bool bResult = true;
    short tmpdata;

	// read the data
	bResult &= Sblive_readac97( index, &tmpdata );

    // bitmasked stuff
    tmpdata &= ~mask;
	data &= mask;
	tmpdata |= data;

	// set the data (back to same address)
	bResult &= Sblive_writeac97( index, tmpdata );

	// return success code
	if( bResult )
	{
		return true;
	}
	else
	{
		return false;
	}

}

/////////////////////////////////////////////////////////////////////////////
// Trace : Write out a line of trace text to the debug file

bool Ac97::Trace( char * pString )
{
    FILE *pFile;

    // open the trace file for append
    pFile = fopen( DEBUG_FILE, "a+" );

    // NOT open OK?
    if( !pFile )
    {
        return false;
    }

    // write out text
    fputs( pString, pFile );

    // close trace file
    fclose( pFile );

    return true;
}

/////////////////////////////////////////////////////////////////////////////
// TraceClear : Delete the trace file

bool Ac97::TraceClear( void )
{
    // kill any trace file
    DeleteFile( DEBUG_FILE );

    return true;
}

// Ac97.cpp : SBLive! Codec Class Implementation File
/////////////////////////////////////////////////////////////////////////////


